package ch.epfl.lara.rmf;

import ch.epfl.lara.rmf.interpretation._;
import ch.epfl.lara.rmf.exsolver.ExSearch;

trait InterpretationGenerator
{
  def interpretations:Stream[Interpretation]
}
                                             
object RMFSolver {
  
  type InterGen = ch.epfl.lara.rmf.exsolver.ExSearch
  
  def main(args: Array[String]) {
//    var density = getModelDensity(5, "../FProblems","..","density.p");
//    Console.println("Considered Problem has a model density of " + density);
      optimizeSpeed();
/*  
      var tries:Int = 100;
      var total:int = 0;
      var max = 0;
      var min = 99999999;
      while (tries > 0) {
        tries -= 1;
        var thisTry =  solveRandSearch(3, 10000000, "../FProblems","..","simple.p");
        if (thisTry < min) {
          min = thisTry;
        }
        if (thisTry > max) {
          max = thisTry;
        }
        total += thisTry;
      }

      Console.println("\t" + (total / 100) + "\taverage iterations");
      Console.println("\t" + (min) + "\tmin iterations");
      Console.println("\t" + (max) + "\tmax iterations");

//      solveRandSearch(4, 10000000, "../FProblems","..","simple.p") match {
      solveExSearch(3, "../FProblems","..","simple.p") match {
        case Some(env) =>  
          Console.println("Model Found");
          env.export foreach (f => RMFPrettyPrinter(f, System.out, 160))
        case None => Console.println("No model found")
    }
*/    
  }

  /** Perform an optimization of speed deltas */
  def optimizeSpeed() {
    //optimization variables
//    var optMax = 0;
//    var optMin = 0;
    var optDeltaI = 0.0;
    var optDeltaD = 0.0;
    var optIterations = 99999999;
    
    //change stack
    var varIterators = new IterableVariable(9) :: new IterableVariable(9) :: Nil;
    var iteratorStack = new IteratorStack(varIterators(0), varIterators(1));

    while (iteratorStack.hasNext()) {
      Speed.deltaI = varIterators(0).getValue() / 10.0;      
      Speed.deltaD = varIterators(0).getValue() / 10.0;      
      
      var tries:Int = 500;
      var total:int = 0;
      while (tries > 0) {
        tries -= 1;
        var thisTry =  solveRandSearch(3, 10000000, "../FProblems","..","simple.p");
        total += thisTry;
      }

      Console.println("\t" + (total / 500) + "\taverage iterations");
      if (optIterations > (total / 500)) {
        optIterations = total / 500;        
        optDeltaI = Speed.deltaI;
        optDeltaD = Speed.deltaD;
      }
      iteratorStack.next();
    }
    
    Console.println("----------");
    Console.println("Optimal Speed: inc " + optDeltaI + " dec " + optDeltaD + " resulting in " + optIterations + " iterations");
  }

  /**
   * Return the density of models (# of models / # of possible interpretations
   */
  def getModelDensity(dSize:Int, pbDir:String, incDir:String, pbFile:String):Double = {
    val input = TPTPParser.parse(pbDir, incDir, pbFile)

    var ts:long = java.lang.System.currentTimeMillis();
    
    //create a list of interpretable function symbols        
    val reprs = ObjectGatherer(Domain(dSize)).gatherObjects(input)

    //create the environment
    var env = new Environment(Domain(dSize), reprs); 
        
    //test: count how many possible interpretations we have in our system
    var counter:long = 0;
    var models:long = 0;
    while (env.hasNext()) {
      counter += 1;
      env.next();
      
      if (Evaluator.evaluate(input, env) == 0) {
        models += 1;
      }
      
      //some visual feedback
      if (counter % 10000 == 0) {
        Console.print(".");
      }
      
      if (counter % 250000 == 0) {
        Console.println();
      }
    }
    
    Console.println();
    var timeTaken:long = (java.lang.System.currentTimeMillis() - ts);
    Console.println("( Domain " + dSize + " , " +  counter + " interpretations considered, " + models + " models found, " + timeTaken + "ms used)");
    return (models.asInstanceOf[Double] / counter.asInstanceOf[Double]);
  }


  /**
   * Solve with exhaustive search
   */
  def solveExSearch(dSize:Int, pbDir:String, incDir:String, pbFile:String):Option[Environment] = {
    val input = TPTPParser.parse(pbDir, incDir, pbFile)

    var ts:long = java.lang.System.currentTimeMillis();
    
    //create a list of interpretable function symbols        
    val reprs = ObjectGatherer(Domain(dSize)).gatherObjects(input)

    //create the environment
    var env = new Environment(Domain(dSize), reprs); 
        
    //test: count how many possible interpretations we have in our system
    var counter:int = 0;
    var found:boolean = false;   
    while (env.hasNext() && found == false) {
      counter += 1;
      env.next();
      found = (Evaluator.evaluate(input, env) == 0);
      
      //some visual feedback
      if (counter % 10000 == 0) {
        Console.print(".");
      }
      
      if (counter % 250000 == 0) {
        Console.println();
      }
    }
    
    Console.println();
    var timeTaken:long = (java.lang.System.currentTimeMillis() - ts);
    Console.println("( Domain " + dSize + " , " +  counter + " interpretations considered, " + timeTaken + "ms used)");
  
    //Return
    if (found) {
      Some(env); 
    } else {  
      None
    }
  }
  
  //-------------------------------------------------------------------------------------
  
  /**
   * Solve with exhaustive search
   */
  def solveRandSearch(dSize:Int, maxIter:Int, pbDir:String, incDir:String, pbFile:String):Int /*Option[Environment]*/ = {
    val input = TPTPParser.parse(pbDir, incDir, pbFile)
    var ts:long = java.lang.System.currentTimeMillis();
      
    //create a list of interpretable function symbols        
    val reprs = ObjectGatherer(Domain(dSize)).gatherObjects(input)

    //create the environment
    var env = new Environment(Domain(dSize), reprs); 
          
    //test: count how many possible interpretations we have in our system
    var counter:Int = 0;
    var found:boolean = false; 
    var rand:java.util.Random = new java.util.Random();
    while (counter < maxIter && found == false) {
      counter += 1;
      env.rand(rand);
      found = (Evaluator.evaluate(input, env) == 0);
        
      //some visual feedback
      if (counter % 10000 == 0) {
        Console.print(".");
      }
        
      if (counter % 250000 == 0) {
        Console.println();
      }
    }
      
    Console.println();
    var timeTaken:long = (java.lang.System.currentTimeMillis() - ts);
    Console.println("( Domain " + dSize + " , " +  counter + " interpretations considered, " + timeTaken + "ms used)");
    
    return counter;
    //Return
//    if (found) {
//      Some(env); 
//    } else {  
//      None
//    }
  }

  //-------------------------------------------------------------------------------------
  
  def displayFuntable(table: Pair[InterpretationElementRepr,Pair[List[List[Element]],List[Any]]]):Unit = {
    table._1 match
    {
      case FunctionRepr(domain, name, n) => Console.println("fct: " + name + "[" + n + "]" )
      case PredicateRepr(domain, name, n) => Console.println("pred: " + name + "[" + n + "]" )
    }
    val PAR = (table._2._1)
    val VAL = (table._2._2)
    val ZIP = (PAR zip VAL)
    ZIP foreach (z => z match
    {
      case Pair(params, value) => 
      {
        params foreach (p => Console.print(" " + p))
        Console.println ("  ---> " + value)
      }
    })
  }
  
  
}
