package ch.epfl.lara.rmf;

/***
 * The TPTP AST
 */

abstract trait TPTPExpression

/*** 
 * Annotations
 */ 

abstract class AnnotedElement(name:String, role:FormulaRole) extends TPTPExpression
case class AnnotedClause(n:String,r:FormulaRole,clause:Clause) extends AnnotedElement(n,r)
case class AnnotedFormula(n:String,r:FormulaRole,formula:Formula) extends AnnotedElement(n,r)

/*** 
 * Formula
 */ 
abstract class Formula extends TPTPExpression
case class Atomic(predicate:String, args: List[Term]) extends Formula
case class Negation(f:Formula) extends Formula
case class And(left:Formula,right:Formula) extends Formula
case class Or(left:Formula,right:Formula) extends Formula
case class Equivalence(left:Formula,right:Formula) extends Formula
case class Implication(left:Formula,right:Formula) extends Formula
case class ReverseImplication(left:Formula,right:Formula) extends Formula
case class Disequivalence(left:Formula,right:Formula) extends Formula
case class NotOr(left:Formula,right:Formula) extends Formula
case class NotAnd(left:Formula,right:Formula) extends Formula
case class Exists(variable: String, f:Formula) extends Formula
case class ForAll(variable: String, f:Formula) extends Formula
case class Term(symbole:Symbol,args:List[Term]) extends TPTPExpression

/*** 
 * Clause
 */ 
case class Clause(literals:List[Literal]) extends TPTPExpression
case class Literal(positive:Boolean, atome:Atomic) extends TPTPExpression
case class Symbol(text:String,variable:Boolean) extends TPTPExpression

/*** 
 * FormulaRole
 */ 
abstract class FormulaRole extends TPTPExpression

case object Axiom extends FormulaRole
case object Hypothesis extends FormulaRole
case object Definition extends FormulaRole 
case object Lemma extends FormulaRole
case object Theorem extends FormulaRole
case object Conjecture extends FormulaRole
case object LemmaConjecture extends FormulaRole
case object NegatedConjecture extends FormulaRole
case object Plain extends FormulaRole
case object FiDomain extends FormulaRole
case object FiFunctors extends FormulaRole
case object FiPredicates extends FormulaRole
case object Unknown extends FormulaRole



