package ch.epfl.lara.rmf;
import ch.epfl.lara.tptp._;


/***
 * This object is a wrapper for the ANTLR Generated Parser (to convert the Java AST to the Scala AST
 */
   
object TPTPParser {
 
  
  def parse(problemDir:String, includeDir:String,problemName:String):List[TPTPExpression] = buildIterable[SimpleTptpParserOutput.TopLevelItem,TPTPExpression](ch.epfl.lara.rmf.TPTPParserTrial.parseTPTPFile(problemDir,includeDir,problemName),build)
   
  
                              
        def build(item:SimpleTptpParserOutput.TopLevelItem):TPTPExpression = item match
        {
                case v: SimpleTptpParserOutput.AnnotatedFormula => AnnotedFormula(v.getName,build(v.getRole().toString()),build(v.getFormula()))
                case v: SimpleTptpParserOutput.AnnotatedClause => AnnotedClause(v.getName,build(v.getRole().toString()),build(v.getClause()))
        }
	
        def build(formula:SimpleTptpParserOutput.Formula):Formula = formula match
        {
                case v: SimpleTptpParserOutput.Formula.Atomic => Atomic(v.getPredicate(),buildListOfTerms(v.getArguments()))
                case v: SimpleTptpParserOutput.Formula.Negation => Negation(build(v.getArgument()))
                case v: SimpleTptpParserOutput.Formula.Binary =>
                {
                  val l = build(v.getLhs())
                  val r = build(v.getRhs())
                  v.getConnective().toString() match
                  {
                    case "&" => And(l,r)
                    case "|" => Or(l,r)
                    case "<=>" => Equivalence(l,r)
                    case "=>" => Implication(l,r)
                    case "<=" =>ReverseImplication(l,r)
                    case "<~>" => Disequivalence(l,r)
                    case "~|" => NotOr(l,r)
                    case "~&" => NotAnd(l,r)
                  }
                }
                case v: SimpleTptpParserOutput.Formula.Quantified => 
                {
                  val f = build(v.getMatrix())
                  v.getQuantifier().toString() match
                  {
                    case "!" => ForAll(v.getVariable(),f)
                    case "?" => Exists(v.getVariable(),f)
                  }
                }
        }
        
        def build(clause: SimpleTptpParserOutput.Clause):Clause = Clause(buildListOfLiterals(clause.getLiterals()))
        def build(term: SimpleTptpParserOutput.Term):Term = Term(build(term.getTopSymbol()),buildListOfTerms(term.getArguments()))
	def build(symbol: SimpleTptpParserOutput.Symbol):Symbol = Symbol(symbol.getText(),symbol.isVariable())
	def build(literal: SimpleTptpParserOutput.Literal):Literal = Literal(literal.isPositive(),build(literal.getAtom()).asInstanceOf[Atomic])
        def build(roleName:String):FormulaRole = roleName match
        {
                case "axiom" => Axiom  
                case "hypothesis" => Hypothesis 
                case "definition" => Definition
                case "lemma" => Lemma
                case  "theorem" => Theorem
                case "conjecture" => Conjecture
                case "lemma_conjecture" => LemmaConjecture
                case "negated_conjecture" => NegatedConjecture
                case "plain" => Plain
                case "fi_domain" => FiDomain
                case "fi_functors" => FiFunctors
                case "fi_predicates" => FiPredicates
                case "unknown" => Unknown
        }
        def buildListOfTerms(iter:java.lang.Iterable):List[Term] = buildIterable(iter,build)
        
        def buildListOfLiterals(iter:java.lang.Iterable):List[Literal] = buildIterable(iter,build)
        
        def buildIterable[J,S](iter:java.lang.Iterable, f: (J => S)):List[S] = if(iter==null) Nil else apply(iter.iterator(),f)
	def apply[J,S](iter:java.util.Iterator, f:(J => S)):List[S] =
        {
   		if(!iter.hasNext()) Nil
                   else
                     iter.next() match
                     {
                     case v:J => f(v)::apply(iter,f)
                     
                     }
        }
                                                          
}
