package ch.epfl.lara.rmf.exsolver;
import ch.epfl.lara.rmf.interpretation._;

case class ExSearch(repr: InterpretationRepr) extends InterpretationGenerator {
  
  override def interpretations:Stream[Interpretation] =
    Stream.fromIterator((enumerateSearchSpace map (i=>buildInterpretation(i))).elements)

  /***
   * Gives all possible experiments: draw k elements out of d, with repetition
   * Iterator has d^k lists of size k
   */
  def enumerateElements(d:Int,k:Int):Iterator[List[Element]] =
    if(k==0) Iterator.single(Nil)
    else if (k==1) Domain(d).elements  map (a => a::Nil)
    else Domain(d).elements flatMap (a => (enumerateElements(d,k-1) map (l => a::l)))
  
  /***
   * Gives all possible experiments: draw 3 elements out of {true, false}, with repetition
   * Iterator has 2^k lists of size k
   */
   def enumerateBooleans(k:Int):Iterator[List[Boolean]] =
      if(k==0) Iterator.single(Nil)
      else if (k==1) (true::false::Nil).elements  map (a => a::Nil)
      else (true::false::Nil).elements flatMap (a => (enumerateBooleans(k-1) map (l => a::l)))
  
   /***
    * Given a list of list (build by enumerateX)
    * Generates a list of the permuations.
    * ex: lists = { { 1 2 3}, { A B C}}
    * gives: { {1 A}, {1 B}, {1 C}, {2 A}, {2 B}, {2 C}, {3 A}, {3 B}, {3 C}}
    */
   def enumerateList[A](iters: List[List[A]]): List[List[A]] = iters match
   {
   case Nil => Nil
   case first::Nil=> first map (a => a::Nil) 
   case first::rest => first flatMap ((a:A) => (enumerateList(rest) map (l => a::l)))
   
   
   }
   
    def buildInterpretation(spec:List[Pair[InterpretationElementRepr,Pair[List[List[Element]],List[Any]]]]):Interpretation =
    {
      val inter = Interpretation(repr.dSize)
      spec foreach (s => s match
          {
        case Pair(FunctionRepr(domain, name,n),Pair(params,values))=> 
        {
          val func = inter.findFunction(name)
          func.asInstanceOf[CustomFunction].setValues(params zip values.asInstanceOf[List[Element]])
        }
        case Pair(PredicateRepr(domain, name,n),Pair(params,values))=> 
        {
          val pred = inter.findPredicate(name)
          pred.asInstanceOf[CustomPredicate].setValues(params zip values.asInstanceOf[List[Boolean]])
        }
          })
      inter
    }
    
    type ParamsComb = List[Element]
   type Params = List[ParamsComb]
   type Values = List[Any]
   type Mapping = Pair[Params,Values]
   type Relation = Pair[InterpretationElementRepr,Mapping]
   type InterCandidate = List[Relation]
   type Decl = Triple[InterpretationElementRepr,Params,List[Values]]
  def enumerateSearchSpace : Stream[InterCandidate] =
    {
      
        
        //Triple 
        // Relation |InterpretationElementRepr|
        // Params Combinations |List[List[Element]]|
        // Possible Values |List[List[Boolean / Element]]|
        
         val fcts = repr.fcts map (f => 
        {
          val paramsComb:Params = enumerateElements(repr.dSize,f.num).toList
          val values:List[Values] = enumerateElements(repr.dSize,paramsComb.length).toList
          Triple(f,paramsComb,values)
        })
        val preds = repr.preds map (p => 
        {
          val paramsComb:Params = enumerateElements(repr.dSize,p.num).toList
          val values:List[Values] = enumerateBooleans(paramsComb.length).toList
          Triple(p,paramsComb,values)
        })
        
        val reprs = (fcts:::preds)
        
        //combine values (as a single list for enumeration)
        val values:List[List[Values]] = reprs map (t => t match
            {
            case Triple(x,p:Params,v:List[Values]) => v
            })
        
      //  val expandedValues :List[List[Values]]  = enumerateList[Values](values)
       //val expandedValues: Stream[List[Values]]=Test.enumerateFromLists[Values](values)
        val expandedValues: Stream[List[Values]] = Test2.enumerateFromLists[Values](values)
       val a =  expandedValues.size
       val b = reprs.size
       
       expandedValues map ((v:List[Values]) =>
       {
         val maps:List[Pair[Values,Decl]] = (v zip reprs);
         maps map ( r=>bind(r))
       })
    }
    

  def bind(data: Pair[Values,Decl]):Relation = data match
  {
    case Pair(values, Triple(x,params,vals)) => Pair(x,Pair(params,values))
  }
                                                                        
  
}
