package ch.epfl.lara.rmf.interpretation;

/***
 * A domain is a set of element. His size is finite.
 */
   
case class Domain(size:Int)
{
  /***
   * Export this domain as TPTP Annoted Formula
   */
  def export:AnnotedFormula = AnnotedFormula("domain",FiDomain,ForAll("X",buildDisjunction(size)))
      
 def buildDisjunction(size: Int):Formula = 
   if(size == 1)
     buildAtom(size)
   else
     Or(buildDisjunction(size-1),buildAtom(size))
     
  def buildAtom(id:Int):Formula = Atomic("=",Term(Symbol("X",true),Nil)::Element(id).export::Nil)
  
  def elements:Iterator[Element] = Iterator.range(1,size+1) map Element
}

/***
 * Element are identified by their ID
 * Scala implementation of case class make sure that: e1 == e2 <=> e1.id == e2.id
 */
   
case class Element(id:Int)
{
  override def toString:String = "E"+id
  
  /***
   * Convenient method to generate the TPTP Term from this element
   */
  def export:Term = Term(Symbol(toString,false),Nil)
}