package ch.epfl.lara.rmf.interpretation;
import ch.epfl.lara.rmf.util.Hypercube;

/***
 * A function is a symbol which maps a list of elements (the parameters) to another element
 * Function are expected to be total.
 */
trait Function extends (List[Element] => Option[Element])
{
  /***
   * Export this function as TPTP Annoted Formula
   */
  def export:AnnotedFormula
}

/***
 * A class describing a function. Mappings params => value have to be "found"
 */
case class CustomFunction(name:String) extends Hypercube[Element,Element] with Function
{
  override def export:AnnotedFormula = AnnotedFormula(name,FiFunctors,buildConjunction(getDefinedEntries))
  
  def buildConjunction(entries: List[Pair[List[Element],Element]]):Formula =  entries match
  {
  case Pair(params,value)::Nil => buildAtom(params,value)
  case Pair(params,value)::rest => And(buildAtom(params,value),buildConjunction(rest))
  case Nil => throw new IllegalArgumentException()
  }
  
  def buildAtom(params:List[Element], value:Element):Formula = Atomic("=",Term(Symbol(name,false),(params map (p=> p.export)))::value.export::Nil)
}
