package ch.epfl.lara.rmf.interpretation;
import ch.epfl.lara.rmf.util.Hypercube;
import scala.collection.mutable.HashMap;

/***
 * An interpretation is a domain and a list of predicates / functions on this domain
 */

case class Interpretation(n:Int)
{
  val domain:Domain = Domain(n)
  val predicates:HashMap[String, Predicate] = new HashMap[String, Predicate]()
  val functions:HashMap[String, Function] = new HashMap[String, Function]()
  
  /***
   * Look for a symbol named "name" in the functions table. Build a new function if
   * the symbol is not found
   */
  def findFunction(name:String):Function = functions.get(name) match
  {
    case Some(f) => f
    case None =>
    {
      val f = CustomFunction(name)
      functions.update(name,f)
      f
    }
  }

  /***
   * Look for a symbol named "name" in the predicates table. Build a new predicate if
   * the symbol is not found. 
   */
  def findPredicate(name:String):Predicate = predicates.get(name) match
  {
    // "=" is a built in predicate.
    case _ if name == "=" => EqualPredicate
    case Some(p) => p
    case None =>
    {
      val p = CustomPredicate(name)
      predicates.update(name,p)
      p
    }
  }

  /***
   * Export this interpretation as a set of TPTP AnnotedFormula
   */
  def export:List[AnnotedFormula]=domain.export::(functions.values map (f => f.export)).toList:::(predicates.values map (p => p.export)).toList
  
}







