package ch.epfl.lara.rmf.interpretation;

/**
 * This class controls two iterators to iterate together. If the second iterator (sub) is null, the
 * iterator will behave the same as self alone
 */
class IteratorStack(self:Iterable, sub:Iterable) extends Iterable {
  
  /** Can we keep iterating? */	  
  override def hasNext():Boolean = {
    if (sub == null) {
      return self.hasNext();
    } else {
      //if either our own iterator or the child still has elements with can continue to iterate
      return self.hasNext() || sub.hasNext();
    }
  }

  /** Iterate to the next possible value */
  override def next():Unit = {
    if (sub == null) {
      return self.next();       
    } else {
      if (sub.hasNext() == false) {
        sub.reset();
        self.next();
      } else {
        sub.next();
      }
    }
  }
  
  /** Reset the iterator */
  override def reset():Unit = {
    if (sub == null) {
      self.reset();
    } else {
      self.reset();
      sub.reset();
    }
  }
  
  def rand(rand:java.util.Random) = {
    if (sub == null) {
      self.rand(rand); 
    } else {
      self.rand(rand);
      sub.rand(rand);
    }
  }
    
}
