package ch.epfl.lara.rmf.interpretation;

/**
 * Gather the predicates and functions within our formula */
case class ObjectGatherer(domain:Domain) {
  
  def gatherObjects(exprs:List[TPTPExpression]):List[InterpretationElementRepr] = exprs match
  {
    case expr::rest => (gatherObjects(expr):::gatherObjects(rest)).removeDuplicates
    case Nil => Nil
  }
  
  def gatherObjects(exprs: TPTPExpression):List[InterpretationElementRepr] = exprs match {
        case AnnotedClause(name, role, clause) => gatherObjects(clause)
        case AnnotedFormula(name, role, formula) => gatherObjects(formula)
        case Atomic(predName, args) if (predName == "=") => gatherObjects(args)
        case Atomic(predName, args) if (predName != "=") => PredicateRepr(domain, predName,args.size)::gatherObjects(args)
        case Negation(f) => gatherObjects(f)
        case And(l, r) => gatherObjects(l::r::Nil)
        case Or(l, r) => gatherObjects(l::r::Nil)
        case Equivalence(l, r) => gatherObjects(l::r::Nil)
        case Implication(l, r) => gatherObjects(l::r::Nil)
        case ReverseImplication(l, r) => gatherObjects(l::r::Nil)
        case Disequivalence(l, r) => gatherObjects(l::r::Nil)
        case NotOr(l, r) => gatherObjects(l::r::Nil)
        case NotAnd(l, r) => gatherObjects(l::r::Nil)
        case Exists(varName, f) => gatherObjects(f)
        case ForAll(varName, f) => gatherObjects(f)
        case Term(Symbol(name, isVar), args) if(isVar) =>  gatherObjects(args)
        case Term(Symbol(name, isVar), args) if(!isVar) => FunctionRepr(domain, name,args.size)::gatherObjects(args)
        case Clause(lits) => gatherObjects(lits)
        case Literal(isPos, atome) => gatherObjects(atome)
  }
  
}
