package ch.epfl.lara.rmf.interpretation;
import ch.epfl.lara.rmf.util.Hypercube;

/***
 * A predicate is a symbole which maps a list of elements (the parameters) to a boolean
 * Predicate are expected to be total.
 */
trait Predicate extends (List[Element] => Option[Boolean])
{
  /***
   * Export this predicate as TPTP Annoted Formula
   */
  def export:AnnotedFormula
}

/***
 * A class describing a predicate. Mappings params => boolean have to be "found"
 */
case class CustomPredicate(name:String) extends Hypercube[Element,Boolean] with Predicate
{
  override def export:AnnotedFormula = AnnotedFormula(name,FiPredicates,buildConjunction(getDefinedEntries))
  
  def buildConjunction(entries: List[Pair[List[Element],Boolean]]):Formula =  entries match
  {
    case Pair(params,value)::Nil => buildAtom(params,value)
    case Pair(params,value)::rest => And(buildAtom(params,value),buildConjunction(rest))
    case Nil => throw new IllegalArgumentException()
  }
  
  def buildAtom(params:List[Element], value:Boolean):Formula = if(value)Atomic(name,(params map (p=> p.export))) else Negation(Atomic(name,(params map (p=> p.export))))
}

 /***
  * The built-in "=" predicate
  */
case object EqualPredicate extends Predicate
{
  override def apply(params:List[Element]):Option[Boolean] = params match
  {
    case a::b::Nil => Some(a==b)
    case _ => throw new IllegalArgumentException()
  }
  
  def export:AnnotedFormula = throw new UnsupportedOperationException()
}
