package ch.epfl.lara.rmf.interpretation;

object TPTPEvaluator {

  type Context = List[Pair[String,Element]]
                      
  def evaluate(as: List[TPTPExpression], inter:Interpretation):Boolean = as forall ( a => evaluate(a,inter))
                      
  def evaluate(a: TPTPExpression, inter: Interpretation): Boolean = a match
  {
    case AnnotedFormula(name,role,f) => evaluate(f,inter,Nil)
    case AnnotedClause(name,role,f) => throw new UnsupportedOperationException()
    case _ => throw new IllegalArgumentException("AnnotedElement required")
  }
                      
  def evaluate(f: Formula, inter: Interpretation, ctx: Context):Boolean = f match
  {
  case Negation(f) => !(evaluate(f,inter,ctx))
  case And(l, r) => evaluate(l,inter,ctx) && evaluate(r,inter,ctx)
  case Or(l, r) =>evaluate(l,inter,ctx) || evaluate(r,inter,ctx)
  
  case Atomic(predName, args) => inter.findPredicate(predName)(args map (a => evaluate(a, inter, ctx))) match
  {
    case Some(result) => result
    // Pred is not complete.
    case None => false
  }
  
  case Implication(l, r) => evaluate(Or(Negation(l),r),inter,ctx)
  case Equivalence(l, r) => evaluate(And(Implication(l,r),Implication(r,l)),inter,ctx)
  case ReverseImplication(l, r) => evaluate(Implication(r,l),inter,ctx)
  case Disequivalence(l, r) => !evaluate(Equivalence(l,r),inter,ctx)
  case NotOr(l, r) => !(evaluate(Or(l,r),inter,ctx))
  case NotAnd(l, r) => !(evaluate(And(l,r),inter,ctx))
  case Exists(varName, f) => inter.domain.elements exists (e => evaluate(f,inter,Pair(varName,e)::ctx))
  case ForAll(varName, f) => inter.domain.elements forall (e => evaluate(f,inter,Pair(varName,e)::ctx))
  }
  
  def evaluate(t: Term, inter: Interpretation, ctx: Context):Element = t match
  {
    case Term(Symbol(name,isVar),args) => 
    if(isVar) // looking in ctx
    {
      ctx find ( c => c._1 == name) match
      {
      case Some(Pair(varName,element)) => element
      case None => throw new RuntimeException("Cannot find var " + name + " in context.")
      }
    }
    else // looking for a function
    {
      inter.findFunction(name)(args map (a => evaluate(a, inter, ctx))) match
      {
        case Some(result) => result
        // Pred is not complete.
        case None => Element(1)
      }
    }
  }
  	
}
